from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "blends.zip",down_path)


bl_info = {
    "name": "Physical Open Waters - POW",
    "author": "Physical Addons",
    "version": (1, 2, 0),
    "blender": (3, 6, 0),
    "description": "Physical Open Waters",
    "location": "World > Atmosphere",
    "doc_url": "https://www.physicaladdons.com/pow/",
    "support": "COMMUNITY",
    "category": "Lighting",
    "tracker_url": "https://discord.gg/wvzPVzj9Vr"
    }

import bpy
from bpy.props import (
    PointerProperty,
)
from bpy.app.handlers import persistent

from . properties import (
    POW_Preset,
    POW_Settings,

    #POW_Ocean,
    register_pow_property_groups,
    update_preset_versions,
    create_preset,
    apply_preset,
)

from . ui import *

from . ocean import (
    POW_OT_AddOcean,
    POW_OT_RemoveOcean,

    setup_default_presets,
)

from . preferences import *


# All the classes that need to be registered in Blender for the addon to work.
# This includes UI classes, PropertyGroups and operators.
REGISTER_CLASSES = (
    # Prop classes
    POW_Preset,
    POW_Settings,
    #POW_Ocean,
    # Operators
    POW_OT_AddOcean,
    POW_OT_RemoveOcean,
    POW_OT_OpenPresetsFolder,
    POW_OT_DeleteCurrentPreset,
	POW_OT_CreateNewPreset,
	POW_OT_SaveCurrentPreset,
    POW_OT_ApplyUnderwaterAll,
    POW_OT_ApplyUnderwaterSelected,
    POW_OT_RemoveUnderwaterAll,
    POW_OT_RemoveUnderwaterSelected,
    POW_OT_AddIntersection,
    POW_OT_RemoveIntersection,
    # UI classes
    POW_PT_OceanSettingsTB,
    POW_PT_MeshSettingsSettingsTB,
    POW_PT_MaterialSettingsTB,
    POW_PT_CameraSettingsTB,
    POW_PT_WavesSettingsTB,
    POW_PT_CapillarySettingsTB,
    POW_PT_WakeSettingsTB,
    POW_PT_UnderwaterSettingsTB,
    POW_PT_FooterTB,
    POW_UL_PresetList,
    # Preferences
    POW_MT_AddonPreferences,
)


@persistent
def depsgraph_update_handler(scene, depsgraph):
    world = scene.world

    selected_materiables = 0
    if len(bpy.context.selected_objects) > 0:
        for object in bpy.context.selected_objects:
            if hasattr(object.data, "materials") and len(object.data.materials) > 0:
                selected_materiables += 1

    if selected_materiables <= 0:
        world.pow_settings["submerged_shader"] = False
        world.pow_settings["submerged_advanced_intersection"] = False
    else:
        # Ensure correct value for the checkbox
        world.pow_settings["submerged_shader"] = True
        for object in bpy.context.selected_objects:
            if hasattr(object.data, "materials"):
                for material in object.data.materials:
                    if not "POW_UnderwaterGroups" in material and not "POW_UnderwaterIgnore" in material:
                        world.pow_settings["submerged_shader"] = False
                        break
        # TODO Check performance impact of all this ^
        
        world.pow_settings["submerged_advanced_intersection"] = True
        for object in bpy.context.selected_objects:
            intersection_node = bpy.data.node_groups.get('POW_Get_Intersection')
            if not intersection_node:
                # Node not present in data, no object is using it
                world.pow_settings["submerged_advanced_intersection"] = False
                break

            if hasattr(object, "modifiers"):
                has_node = False
                for mod in object.modifiers:
                    if mod.type == 'NODES':
                        if mod.node_group == intersection_node:
                            has_node = True
                            break
                if not has_node:
                    world.pow_settings["submerged_advanced_intersection"] = False
                    break


def invalid_preset_index_fix():
    try:
        world = bpy.context.scene.world
        preset = world.pow_settings.presets[world.pow_settings.preset_index]
        print("Preset exists:", preset)
    except IndexError:
        print("[POW] Warning: No preset exists at index:", world.pow_settings.preset_index)
        print("[POW] Defaulting to preset at index 0.")
        world.pow_settings["preset_index"] = 0 # reset preset index


@persistent
def load_handler(file):
    """
    Handler called upon loading a .blend file.
    Version checks are done at this point.
    """
    invalid_preset_index_fix()

    ##### Update ocean to latest POW version
    TEMP_NAME = "TEMP_POW_UPDATE_PRESET_add_tail_to_reduce_accidental_preset_naming_to_this"
    context = bpy.context

    scene = context.scene
    if not scene:
        return # No scene, nothing to update

    world = scene.world
    if not world:
        return # No world, nothing to update

    pow_settings = world.pow_settings
    blend_version = pow_settings.get("blend_version")

    if not pow_settings.ocean:
        return # Only need to update if ocean is present.

    if blend_version != str(get_addon_version()):
        # Create a preset with current ocean data stored
        pow_settings.preset_name = TEMP_NAME
        base_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
        os.makedirs(base_path, exist_ok=True)
        full_path = os.path.join(base_path, world.pow_settings.preset_name + ".json")
        if not os.path.isfile(full_path):
            create_preset(context, full_path)
            update_preset_list(None, context)
            # Set the active preset index to the newly created one
            for i in range(len(world.pow_settings.presets)):
                if world.pow_settings.presets[i].name == world.pow_settings.preset_name:
                    world.pow_settings.preset_index = i
                    break

        bpy.ops.pow.remove_ocean() # Remove old ocean
        bpy.ops.pow.add_ocean() # Add new ocean

        # Apply stored temp preset
        if len(world.pow_settings.presets) > world.pow_settings.preset_index and world.pow_settings.preset_index >= 0:
            preset_data = world.pow_settings.presets[world.pow_settings.preset_index]
            apply_preset(context, preset_data.path) # Apply preset
            world.pow_settings.preset_name = preset_data.name

        # Remove temp preset
        if world.pow_settings.presets[world.pow_settings.preset_index].name == TEMP_NAME:
            world = context.scene.world
            # Find the target directory for our presets and delete it
            base_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
            #os.makedirs(base_path, exist_ok=True)
            full_path = os.path.join(base_path, world.pow_settings.presets[world.pow_settings.preset_index].name + ".json")
            os.remove(full_path)
            # Set the selected preset to default
            world.pow_settings.preset_index = -1
            update_preset_list(None, context)
    #####


def register():
    # Register all classes
    for cls in REGISTER_CLASSES:
        bpy.utils.register_class(cls)
    
    bpy.types.World.pow_settings = PointerProperty(type = POW_Settings)
    register_pow_property_groups(bpy.context, {}, first_time = True)
    #bpy.types.Object.pow_ocean = PointerProperty(type = POW_Ocean)

    # Install default presets in addon's preset folder after addon has been installed
    setup_default_presets()
    update_preset_versions(bpy.context)
    #invalid_preset_index_fix()

    # Handlers
    bpy.app.handlers.depsgraph_update_post.append(depsgraph_update_handler)
    bpy.app.handlers.load_post.append(load_handler)


def unregister():
    prefs = bpy.context.preferences.addons[__package__].preferences
    if prefs.animatable_props:
        bpy.app.handlers.frame_change_post.remove(frame_change_handler)

    # Unregister all classes
    for cls in REGISTER_CLASSES:
        bpy.utils.unregister_class(cls)
    
    del bpy.types.World.pow_settings
    del bpy.types.Object.pow_ocean

    # Handlers
    bpy.app.handlers.depsgraph_update_post.remove(depsgraph_update_handler)
    bpy.app.handlers.load_post.remove(load_handler)


if __name__ == "__main__":
    register()
