import bpy
from bpy.types import(
    Operator,
)

from mathutils import (
    Vector
)

import os
import shutil

from . waves import update_waves

from .preferences import get_addon_version


def link_driver_simple(source, target, property, data_path, index=-1, id_type='OBJECT'):
    # Adds simple driver linking the property on source with target's dataPath
    # Use index > 0 if you need to indicate the target channel of property that the driver affects
    # id_type is the type of target
    d = source.driver_add(property, index).driver

    for var in d.variables:
        d.variables.remove(var)

    v = d.variables.new()
    v.name = "X"
    v.targets[0].id_type = id_type
    v.targets[0].id = target
    v.targets[0].data_path = data_path

    d.expression = v.name


def link_driver_multiply(source, target_1, target_2, property, data_path_1, data_path_2, index=-1, id_type_1='OBJECT', id_type_2='OBJECT'):
    d = source.driver_add(property, index).driver

    for var in d.variables:
        d.variables.remove(var)

    v1 = d.variables.new()
    v1.name = property + "1"
    v1.targets[0].id_type = id_type_1
    v1.targets[0].id = target_1
    v1.targets[0].data_path = data_path_1

    v2 = d.variables.new()
    v2.name = property + "2"
    v2.targets[0].id_type = id_type_2
    v2.targets[0].id = target_2
    v2.targets[0].data_path = data_path_2

    d.expression = v1.name + "*" + v2.name


def update_wakes(self, context):
    world = context.scene.world
    ocean = world.pow_settings.ocean

    ocean.modifiers["PhysicalOpenWaters"]["Input_33"] = world.pow_settings.wake_object

    #material = ocean.material_slots[0].material.node_tree
    # TODO Cleanup the .blend file to make node group names better
    #kelvin_object_node = material.nodes["Group.048"].node_tree.nodes["Texture Coordinate"]
    #kelvin_object_node.object = world.pow_settings.wake_object


def update_shrinkwrap(self, context):
    world = context.scene.world
    ocean = world.pow_settings.ocean

    ocean.modifiers["PhysicalOpenWaters"]["Input_59"] = world.pow_settings.shrinkwrap_object
    ocean.modifiers["PhysicalOpenWaters"]["Input_60"] = (world.pow_settings.shrinkwrap_object != None)



def update_world_volumetrics(self, context):
    world = context.scene.world
    node_tree = world.node_tree
    underwater_fog = node_tree.nodes.get("POW Underwater Fog")
    material_output = node_tree.get_output_node('ALL')
    ocean = world.pow_settings.ocean

    # TODO Investigate cases where volume is linked and prevent the link from breaking when volume is added
    if world.pow_settings.world_volumetrics == 'enabled':
        # Link volume and add drivers
        node_tree.links.new(underwater_fog.outputs[2], material_output.inputs[1])
        link_driver_simple(underwater_fog.inputs[6], ocean, 'default_value', "pow_ocean.volume_density")
        link_driver_simple(underwater_fog.inputs[7], ocean, 'default_value', "pow_ocean.volume_distance")
    else:
        # Unlink volume and remove drivers (for performance)
        node_tree.links.remove(material_output.inputs[1].links[0])
        underwater_fog.inputs[6].driver_remove('default_value')
        underwater_fog.inputs[7].driver_remove('default_value')


def add_intersection_nodes(object):
    if object == None or not hasattr(object, "modifiers"):
        return

    intersection_node = bpy.data.node_groups['POW_Get_Intersection']

    has_node = False

    for mod in object.modifiers:
        if mod.type == 'NODES':
            if mod.node_group == intersection_node:
                has_node = True
                break
    
    if not has_node:
        mod = object.modifiers.new("POW_Get_Intersection", 'NODES')
        mod.node_group = intersection_node
        mod["Socket_2"] = bpy.context.scene.world.pow_settings.ocean
        mod["Input_14"] = True

    for mat in object.material_slots:
        material = mat.material
        node_tree = material.node_tree

        # Update materials to use correct data
        for node in node_tree.nodes:
            if node.name == "POW Underwater Attenuation":
                node.inputs[2].default_value = 1.0
            elif node.name == "POW Underwater Fog":
                node.inputs[5].default_value = 1.0
            elif node.name == "POW Surface Tension":
                node.inputs[1].default_value = 1.0


def remove_intersection_nodes(object):
    if object == None or not hasattr(object, "modifiers"):
        return

    intersection_node = bpy.data.node_groups.get('POW_Get_Intersection')

    if intersection_node:
        for mod in object.modifiers:
            if mod.type == 'NODES':
                if mod.node_group == intersection_node:
                    object.modifiers.remove(mod)
        
        for mat in object.material_slots:
            material = mat.material
            if material == None: # sometimes there are materials that are added without name or node_tree. Skip These
                continue
            node_tree = material.node_tree

            # Update materials to use correct data
            for node in node_tree.nodes:
                if node.name == "POW Underwater Attenuation":
                    node.inputs[2].default_value = 0.0
                elif node.name == "POW Underwater Fog":
                    node.inputs[4].default_value = 0.0
                elif node.name == "POW Surface Tension":
                    node.inputs[1].default_value = 0.0


def add_underwater_nodes(material):
    surface_tension = bpy.data.node_groups['POW_Surface_Tension']
    underwater_fog = bpy.data.node_groups['POW_Underwater_Fog']
    underwater_attenuation = bpy.data.node_groups['POW_Underwater_Attentuation']

    if not "POW_UnderwaterGroups" in material and not "POW_UnderwaterIgnore" in material:
        material.use_nodes = True  # enable 'Use nodes' if not enabled already
        node_tree = material.node_tree
        material_output = node_tree.get_output_node('ALL')

        if material_output is None:
            # If there's no output, we can't add our nodes so do nothing
            return
        
        # Check output "surface" connection and get the link
        input = material_output.inputs[0]
        if input != None:
            if input.is_linked:
                link = input.links[0]
                if link.is_valid:
                    pre_node = link.from_node

                    attenuation_node = node_tree.nodes.new(type='ShaderNodeGroup')
                    attenuation_node.name = "POW Underwater Attenuation"
                    attenuation_node.node_tree = underwater_attenuation

                    fog_node = node_tree.nodes.new(type='ShaderNodeGroup')
                    fog_node.name = "POW Underwater Fog"
                    fog_node.node_tree = underwater_fog

                    tension_node = node_tree.nodes.new(type='ShaderNodeGroup')
                    tension_node.name = "POW Surface Tension"
                    tension_node.node_tree = surface_tension

                    node_tree.links.new(tension_node.outputs[0], material_output.inputs[0])
                    node_tree.links.new(fog_node.outputs[0], tension_node.inputs[0])
                    node_tree.links.new(attenuation_node.outputs[0], fog_node.inputs[0])
                    node_tree.links.new(pre_node.outputs[0], attenuation_node.inputs[0])

                    material["POW_UnderwaterGroups"] = True


def remove_underwater_nodes(material):
    if "POW_UnderwaterGroups" in material:
        node_tree = material.node_tree
        material_output = node_tree.get_output_node('ALL')

        attenuation_node = node_tree.nodes.get("POW Underwater Attenuation")
        fog_node = node_tree.nodes.get("POW Underwater Fog")
        tension_node = node_tree.nodes.get("POW Surface Tension")

        input = attenuation_node.inputs[0]
        if input != None:
            if input.is_linked:
                link = input.links[0]
                if link.is_valid:
                    pre_node = link.from_node
                    node_tree.links.new(pre_node.outputs[0], material_output.inputs[0])

                    node_tree.nodes.remove(attenuation_node)
                    node_tree.nodes.remove(fog_node)
                    node_tree.nodes.remove(tension_node)

                    del material["POW_UnderwaterGroups"]


def update_underwater(self, context, add=True, all_materials=False):
    if add:
        if all_materials:
            for material in bpy.data.materials:
                add_underwater_nodes(material)
        else:
            for object in context.selected_objects:
                for material in object.data.materials:
                    add_underwater_nodes(material)
    else:
        if all_materials:
            for object in bpy.data.objects:
                if not object == context.scene.world.pow_settings.ocean:
                    remove_intersection_nodes(object)
            for material in bpy.data.materials:
                remove_underwater_nodes(material)
        else:
            for object in context.selected_objects:
                for material in object.data.materials:
                    remove_underwater_nodes(material)


def update_submerged(self, context):
    world = context.scene.world
    if world.pow_settings.submerged_shader:
        for object in context.selected_objects:
            if hasattr(object.data, "materials"):
                for material in object.data.materials:
                    add_underwater_nodes(material)
    else:
        # Remove advanced intersection first for cleanup
        world.pow_settings.submerged_advanced_intersection = False
        for object in context.selected_objects:
            if hasattr(object.data, "materials"):
                for material in object.data.materials:
                    remove_underwater_nodes(material)


def update_intersection(self, context):
    world = context.scene.world
    if world.pow_settings.submerged_advanced_intersection:
        for object in context.selected_objects:
            add_intersection_nodes(object)
    else:
        for object in context.selected_objects:
            remove_intersection_nodes(object)
            

def add_world_underwater(world):
    node_tree = world.node_tree

    underwater_fog = bpy.data.node_groups['POW_Underwater_Fog']

    material_output = node_tree.get_output_node('ALL')
    if material_output is None:
        # If there's no output, we can't add our nodes so do nothing
        return
    
    # Check output "surface" connection and get the link
    input = material_output.inputs[0]
    if input != None:
        if input.is_linked:
            link = input.links[0]
            if link.is_valid:
                pre_node = link.from_node

                # Assume pre-node is Background or Emission and look for first link.
                color_node = None
                input1 = pre_node.inputs[0]
                if input1 != None and input1.is_linked:
                    link1 = input1.links[0]
                    if link1.is_valid:
                        color_node = link1.from_node

                fog_node = node_tree.nodes.new(type='ShaderNodeGroup')
                fog_node.name = "POW Underwater Fog"
                fog_node.node_tree = underwater_fog

                node_tree.links.new(fog_node.outputs[0], material_output.inputs[0])
                #if color_node:
                #    node_tree.links.new(color_node.outputs[0], fog_node.inputs[1])
                #else:
                #    fog_node.inputs[1].default_value = pre_node.inputs[0].default_value
                node_tree.links.new(pre_node.outputs[0], fog_node.inputs[0])

                # Set the node to "world shader" mode
                fog_node.inputs[3].default_value = 1.0

                world["POW_PreviousShaderNode"] = pre_node.name
                world["POW_WorldUnderwater"] = True


def remove_world_underwater(world):
    node_tree = world.node_tree

    material_output = node_tree.get_output_node('ALL')
    underwater_fog = node_tree.nodes.get("POW Underwater Fog")

    if underwater_fog:
        if world["POW_PreviousShaderNode"]:
            node = node_tree.nodes.get(world["POW_PreviousShaderNode"])
            if node:
                node_tree.links.new(node.outputs[0], material_output.inputs[0])
                del world["POW_PreviousShaderNode"]
        node_tree.nodes.remove(underwater_fog)
        del world["POW_WorldUnderwater"]


def update_world_underwater(self, context):
    world = context.scene.world

    if world.pow_settings.world_underwater == 'enabled':
        if "POW_WorldUnderwater" not in world:
            add_world_underwater(world)
    elif "POW_WorldUnderwater" in world:
        remove_world_underwater(world)


def update_ocean(self, context):
    world = context.scene.world
    ocean = world.pow_settings.ocean

    ocean_node_tree = ocean.material_slots[0].material.node_tree
    # FIXME Hardcoded name is easy to break, needs a better approach
    ocean_modifier = ocean.modifiers["PhysicalOpenWaters"]

    # Material
    #link_driver_simple(ocean_modifier, ocean, '["Input_55"]', "pow_ocean.water_color.r", index=0)
    #link_driver_simple(ocean_modifier, ocean, '["Input_55"]', "pow_ocean.water_color.g", index=1)
    #link_driver_simple(ocean_modifier, ocean, '["Input_55"]', "pow_ocean.water_color.b", index=2)

    #link_driver_simple(ocean_modifier, ocean, '["Input_56"]', "pow_ocean.water_subsurface_scale")

    # Mesh props
    link_driver_simple(ocean_modifier, ocean, '["Input_21"]', "pow_ocean.base_size")
    link_driver_simple(ocean_modifier, ocean, '["Input_23"]', "pow_ocean.base_resolution")
    link_driver_simple(ocean_modifier, ocean, '["Input_25"]', "pow_ocean.detail_resolution")

    # Camera props
    link_driver_simple(ocean_modifier, ocean, '["Input_54"]', "pow_ocean.view_frustum_boost")
    link_driver_simple(ocean_modifier, ocean, '["Input_30"]', "pow_ocean.far_lod_offset")
    link_driver_simple(ocean_modifier, ocean, '["Input_31"]', "pow_ocean.near_lod_offset")
    link_driver_simple(ocean_modifier, ocean, '["Input_32"]', "pow_ocean.lod_exponent")

    # Wave props
    link_driver_simple(ocean_modifier, ocean, '["Input_2"]', "pow_ocean.time_scale")
    link_driver_simple(ocean_modifier, ocean, '["Input_7"]', "pow_ocean.wave_choppiness")
    link_driver_simple(ocean_modifier, ocean, '["Input_11"]', "pow_ocean.wave_skew")
    link_driver_simple(ocean_modifier, ocean, '["Socket_1"]', "pow_ocean.foam_amount")

    # Wake props
    link_driver_simple(ocean_modifier, ocean, '["Input_34"]', "pow_ocean.wake_height")

    # Capillary props
    #link_driver_simple(ocean_modifier, ocean, '["Input_42"]', "pow_ocean.capillary_strength")
    #link_driver_simple(ocean_modifier, ocean, '["Input_43"]', "pow_ocean.capillary_scale")
    #link_driver_simple(ocean_modifier, ocean, '["Input_44"]', "pow_ocean.capillary_stretch")
    #link_driver_simple(ocean_modifier, ocean, '["Input_45"]', "pow_ocean.capillary_direction")
    #link_driver_simple(ocean_modifier, ocean, '["Input_46"]', "pow_ocean.capillary_speed")
    #link_driver_simple(ocean_modifier, ocean, '["Input_47"]', "pow_ocean.capillary_variance")
    #link_driver_simple(ocean_modifier, ocean, '["Input_48"]', "pow_ocean.capillary_drag")
    #link_driver_simple(ocean_modifier, ocean, '["Input_49"]', "pow_ocean.capillary_mask_scale")
    #link_driver_simple(ocean_modifier, ocean, '["Input_50"]', "pow_ocean.capillary_mask_minimum")
    #link_driver_simple(ocean_modifier, ocean, '["Input_51"]', "pow_ocean.capillary_mask_maximum")
    #link_driver_simple(ocean_modifier, ocean, '["Input_52"]', "pow_ocean.capillary_roughness_minimum")
    #link_driver_simple(ocean_modifier, ocean, '["Input_53"]', "pow_ocean.capillary_roughness_maximum")

    # Particles
    link_driver_simple(ocean_modifier, ocean, '["Input_61"]', "pow_ocean.particle_domain_size")
    link_driver_simple(ocean_modifier, ocean, '["Input_63"]', "pow_ocean.particle_domain_z_offset")

    # Water edge
    link_driver_simple(ocean_modifier, ocean, '["Input_66"]', "pow_ocean.water_edge_z_offset")
    link_driver_simple(ocean_modifier, ocean, '["Input_67"]', "pow_ocean.water_edge_width")
    link_driver_simple(ocean_modifier, ocean, '["Input_68"]', "pow_ocean.water_edge_height")
    link_driver_simple(ocean_modifier, ocean, '["Input_69"]', "pow_ocean.water_edge_resolution")

    # Material settings group
    material_settings = bpy.data.node_groups['POW_Material_settings']

    link_driver_simple(material_settings.nodes["SSS Scale"].outputs[0], ocean, 'default_value', "pow_ocean.water_subsurface_scale")

    link_driver_simple(material_settings.nodes["Reflection Multiply Fresnel"].outputs[0], ocean, 'default_value', "pow_ocean.reflection_multiply_fresnel")
    link_driver_simple(material_settings.nodes["Reflection IOR"].outputs[0], ocean, 'default_value', "pow_ocean.reflection_ior")
    link_driver_simple(material_settings.nodes["Refraction IOR"].outputs[0], ocean, 'default_value', "pow_ocean.refraction_ior")
    link_driver_simple(material_settings.nodes["Refraction Dispersion"].outputs[0], ocean, 'default_value', "pow_ocean.refraction_dispersion")

    link_driver_multiply(material_settings.nodes["Sun Color"].outputs[0], world, world, 'default_value',
        "pow_settings.sun.data.color.r", "pow_settings.sun.data.energy", index=0, id_type_1='WORLD', id_type_2='WORLD')
    link_driver_multiply(material_settings.nodes["Sun Color"].outputs[0], world, world, 'default_value',
        "pow_settings.sun.data.color.g", "pow_settings.sun.data.energy", index=1, id_type_1='WORLD', id_type_2='WORLD')
    link_driver_multiply(material_settings.nodes["Sun Color"].outputs[0], world, world, 'default_value',
        "pow_settings.sun.data.color.b", "pow_settings.sun.data.energy", index=2, id_type_1='WORLD', id_type_2='WORLD')

    link_driver_simple(material_settings.nodes["Sun Vector"].inputs[0], world, 'default_value', "pow_settings.sun.rotation_euler.x", index=0, id_type='WORLD')
    link_driver_simple(material_settings.nodes["Sun Vector"].inputs[0], world, 'default_value', "pow_settings.sun.rotation_euler.y", index=1, id_type='WORLD')
    link_driver_simple(material_settings.nodes["Sun Vector"].inputs[0], world, 'default_value', "pow_settings.sun.rotation_euler.z", index=2, id_type='WORLD')

    link_driver_simple(material_settings.nodes["Extinction Coefficient"].outputs[0], ocean, 'default_value', "pow_ocean.extinction_coefficient.r", index=0)
    link_driver_simple(material_settings.nodes["Extinction Coefficient"].outputs[0], ocean, 'default_value', "pow_ocean.extinction_coefficient.g", index=1)
    link_driver_simple(material_settings.nodes["Extinction Coefficient"].outputs[0], ocean, 'default_value', "pow_ocean.extinction_coefficient.b", index=2)

    # It's 3 not 1 for some weird reason, something to do with indexed array inputs in Blender that I just learnt.
    link_driver_simple(material_settings.nodes["Scattering Distance"].outputs[0], ocean, 'default_value', "pow_ocean.scattering_distance")

    link_driver_simple(material_settings.nodes["Water Color"].outputs[0], ocean, 'default_value', "pow_ocean.water_color.r", index=0)
    link_driver_simple(material_settings.nodes["Water Color"].outputs[0], ocean, 'default_value', "pow_ocean.water_color.g", index=1)
    link_driver_simple(material_settings.nodes["Water Color"].outputs[0], ocean, 'default_value', "pow_ocean.water_color.b", index=2)

    link_driver_simple(material_settings.nodes["Vertical Scattering"].outputs[0], ocean, 'default_value', "pow_ocean.vertical_scattering")
    link_driver_simple(material_settings.nodes["Vertical Scattering Exponent"].outputs[0], ocean, 'default_value', "pow_ocean.vertical_scattering_exponent")
    link_driver_simple(material_settings.nodes["Snells Scattering"].outputs[0], ocean, 'default_value', "pow_ocean.snell_scattering")
    link_driver_simple(material_settings.nodes["Forward Scattering"].outputs[0], ocean, 'default_value', "pow_ocean.forward_scattering")
    link_driver_simple(material_settings.nodes["Back Scattering"].outputs[0], ocean, 'default_value', "pow_ocean.back_scattering")
    link_driver_simple(material_settings.nodes["Camera Near Offset"].outputs[0], ocean, 'default_value', "pow_ocean.camera_near_offset")
    link_driver_simple(material_settings.nodes["Water Height"].outputs[0], ocean, 'default_value', "location.z")

    link_driver_simple(material_settings.nodes["Capillary Strength"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_strength")
    link_driver_simple(material_settings.nodes["Capillary Scale"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_scale")
    link_driver_simple(material_settings.nodes["Capillary Stretch"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_stretch")
    link_driver_simple(material_settings.nodes["Capillary Direction"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_direction")
    link_driver_simple(material_settings.nodes["Capillary Speed"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_speed")
    link_driver_simple(material_settings.nodes["Capillary Variance"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_variance")
    link_driver_simple(material_settings.nodes["Capillary Drag"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_drag")
    link_driver_simple(material_settings.nodes["Noise Mask Scale"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_mask_scale")

    link_driver_simple(material_settings.nodes["Noise Mask Position"].inputs[0], ocean, 'default_value', "pow_ocean.capillary_mask_offset.x", index=0)
    link_driver_simple(material_settings.nodes["Noise Mask Position"].inputs[0], ocean, 'default_value', "pow_ocean.capillary_mask_offset.y", index=1)
    link_driver_simple(material_settings.nodes["Noise Mask Position"].inputs[0], ocean, 'default_value', "pow_ocean.capillary_mask_offset.z", index=2)

    link_driver_simple(material_settings.nodes["Mask Minimum"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_mask_minimum")
    link_driver_simple(material_settings.nodes["Mask Maximum"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_mask_maximum")
    link_driver_simple(material_settings.nodes["Roughness Minimum"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_roughness_minimum")
    link_driver_simple(material_settings.nodes["Roughness Maximum"].outputs[0], ocean, 'default_value', "pow_ocean.capillary_roughness_maximum")

    link_driver_simple(material_settings.nodes["Ripple Wave Height"].outputs[0], ocean, 'default_value', "pow_ocean.ripple_wave_height")
    link_driver_simple(material_settings.nodes["Ripple Time Scale"].outputs[0], ocean, 'default_value', "pow_ocean.ripple_time_scale")
    link_driver_simple(material_settings.nodes["Ripple Frequency Minimum"].outputs[0], ocean, 'default_value', "pow_ocean.ripple_frequency_minimum")
    link_driver_simple(material_settings.nodes["Ripple Frequency Maximum"].outputs[0], ocean, 'default_value', "pow_ocean.ripple_frequency_maximum")
    link_driver_simple(material_settings.nodes["Ripple Fade Distance"].outputs[0], ocean, 'default_value', "pow_ocean.ripple_fade_distance")

    # Update toggles
    material_settings.nodes["SSS/Refraction"].outputs[0].default_value = ocean.pow_ocean.material_type == 'refractive'

    ocean_modifier["Input_60"] = world.pow_settings.shrinkwrap_object != None
    ocean_modifier["Input_64"] = ocean.pow_ocean.particles == 'enabled'
    ocean_modifier["Input_74"] = ocean.pow_ocean.water_edge == 'enabled'

#    ocean_modifier["Input_21"] = ocean.pow_ocean.base_size
#    ocean_modifier["Input_23"] = ocean.pow_ocean.base_resolution
#    ocean_modifier["Input_25"] = ocean.pow_ocean.detail_resolution
#
#    ocean.data.update() # Required for geometry nodes to register change in inputs

    update_waves(None, context)


def update_camera(self, context):
    world = context.scene.world
    camera = world.pow_settings.camera
    ocean = world.pow_settings.ocean

    if camera and ocean:

        ocean.modifiers["PhysicalOpenWaters"]["Input_24"] = camera # Set "Camera" prop
        node_group = ocean.modifiers["PhysicalOpenWaters"].node_group
        frustum_tree = node_group.nodes["Group.003"].node_tree
        store_attributes_tree = node_group.nodes["Group.015"].node_tree

        ocean.data.update() # Update ocean object data to display changes in geometry nodes

        link_driver_simple(frustum_tree.nodes["Value.003"].outputs[0], camera.data, "default_value", "sensor_width", id_type='CAMERA')
        link_driver_simple(frustum_tree.nodes["Value"].outputs[0], camera.data, "default_value", "lens", id_type='CAMERA')

        link_driver_simple(store_attributes_tree.nodes["z_near"].outputs[0], camera.data, "default_value", "clip_start", id_type='CAMERA')
        link_driver_simple(store_attributes_tree.nodes["z_far"].outputs[0], camera.data, "default_value", "clip_end", id_type='CAMERA')


def append_ocean_data():
    # Append the geometry nodes
    path = os.path.join(os.path.dirname(__file__), "blends/ocean.blend")
    specific_objects = ["POW_Ocean"] #,"POW_Cube"]

    with bpy.data.libraries.load(path) as (data_from, data_to):
        data_to.objects = [obj for obj in data_from.objects if obj in specific_objects and obj not in bpy.context.scene.objects]

    for obj in data_to.objects:
        if obj is not None:
            bpy.context.scene.collection.objects.link(obj)
        if obj.name.startswith("POW_Ocean"):
            bpy.context.scene.world.pow_settings.ocean = obj


def setup_default_presets():
    pow_presets_dir = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
    src_presets_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "presets")

    if not os.path.exists(pow_presets_dir):
        os.makedirs(pow_presets_dir)
    for filename in os.listdir(src_presets_dir):
        src_preset_path = os.path.join(src_presets_dir, filename)
        if os.path.isfile(src_preset_path):
            shutil.copy2(src_preset_path, pow_presets_dir)
    return 0


def cleanup_pow_nodes():
    """Remove all POW node groups to avoid duplications."""
    group_name_prefixes = ('POW_', '. POW_')
    for node_group in bpy.data.node_groups:
        if any(node_group.name.startswith(s) for s in group_name_prefixes):
            bpy.data.node_groups.remove(node_group)


class POW_OT_AddOcean(Operator):
    bl_idname = "pow.add_ocean"
    bl_label = "Add Ocean"
    bl_description = "Add ocean to the scene"

    def execute(self, context):
        world = context.scene.world
        append_ocean_data()

        # Set sun
        if world.pow_settings.sun == None:
            if context.scene.objects.get("Starlight Sun"):
                # Use PSA sun by default
                world.pow_settings.sun = context.scene.objects.get("Starlight Sun")
            else:
                # Find brightest sun and use it
                brightest_sun = None
                for obj in context.scene.objects:
                    if obj.type == 'LIGHT' and obj.data.type == 'SUN':
                        if brightest_sun:
                            max_obj = max(obj.data.color.r, obj.data.color.b, obj.data.color.b)
                            max_bri = max(brightest_sun.data.color.r, brightest_sun.data.color.b, brightest_sun.data.color.b)
                            if obj.data.energy * max_obj > brightest_sun.data.energy * max_bri:
                                brightest_sun = obj
                        else:
                            brightest_sun = obj
                world.pow_settings.sun = brightest_sun
            
        
        # Add underwater to world by default
        if "POW_WorldUnderwater" not in world:
            world.pow_settings.world_underwater = 'enabled'


        ocean = world.pow_settings.ocean
        for slot in ocean.material_slots:
            slot.material["POW_UnderwaterIgnore"] = True
        if ocean:
            # FIXME These probably shouldn't be hardcoded name references
            node_group = ocean.modifiers["PhysicalOpenWaters"].node_group
            frustum_tree = node_group.nodes["Group.003"].node_tree

            link_driver_simple(frustum_tree.nodes["Value.001"].outputs[0], bpy.context.scene, "default_value", "render.resolution_x", id_type='SCENE')
            link_driver_simple(frustum_tree.nodes["Value.002"].outputs[0], bpy.context.scene, "default_value", "render.resolution_y", id_type='SCENE')

            world.pow_settings.camera = bpy.context.scene.camera

            ocean.hide_select = True

            # Reset props to defaults
            for name, prop in ocean.pow_ocean.rna_type.properties.items():
                if not name == "rna_type":
                    ocean.pow_ocean.property_unset(name)
        
        world.pow_settings.blend_version = str(get_addon_version())
        
        update_ocean(ocean, context)

        world.pow_settings.show_presets = True # Open preset list

        return {'FINISHED'}


class POW_OT_RemoveOcean(Operator):
    bl_idname = "pow.remove_ocean"
    bl_label = "Remove Ocean"
    bl_description = "Remove ocean from the scene"

    def invoke(self, context, event):
        # Create a confirmation popup to reduce chance of accidental deletion
        return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        world = context.scene.world
        
        ocean = world.pow_settings.ocean

        # Remove all underwater interactions first
        update_underwater(self, context, add=False, all_materials=True)
        world.pow_settings.world_underwater = 'disabled'

        world.pow_settings.ocean = None
        bpy.data.objects.remove(ocean)

        cleanup_pow_nodes()

        return {'FINISHED'}